/*
 * mkdisc.c
 *
 * Make a WORM disc using a Sony CDW-E1 encoding unit
 *
 * Copyright 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <dslib.h>
#include <fcntl.h>
#include <stdio.h>
#include <malloc.h>
#include <unistd.h>
#include <signal.h>

#define CDROM_BLOCKSIZE 2048
#define WSTART 0xe0
#define WCONT 0xe1
#define RBLIMIT 0x05
#define DISCONTINUE 0xe2

#define FIXPROTO (uchar_t *)

#define INTTOTWODIGITBCD( i )      ((((i) / 10) << 4) | ((i) % 10))
#define CDEW1_SENSELEN 22
#define PRINTSENSE(dsp) { \
    int i;\
    if (STATUS(dsp) == STA_CHECK) {\
	for (i = 0; i < CDEW1_SENSELEN; i++)\
	    fprintf(stderr, "%02x ", SENSEBUF(dsp)[i]);\
	fprintf(stderr, "\n");\
    }}

/*
 * The numbers here need to be in BCD
 */
typedef struct {
    unsigned char control, tno, index, dform, zero, min, sec, frame;
} ind_t;

/*
 * Here we make very limited use of the capabilities of the CDW-E1.
 * It can lay down up to 99 tracks.  The purpose of this program
 * is to transfer filesystems to disc, so it only writes one track
 * per disc.
 */
typedef struct {
    ind_t leadin, gap, data, leadout;
} cuesheet_t;

/*
 * Store dsp in here so signal handler can see it if needed
 */
static dsreq_t *sig_dsp;

/*
 * Send the cue sheet to the writer
 *
 * Note that this assumes cue sheet is less than 64K
 */
static void
write_start(dsreq_t *dsp, cuesheet_t *cue, size_t sz)
{
    fillg1cmd(dsp, FIXPROTO CMDBUF(dsp), WSTART,
     0x10, B2(sz), 0, 0, 0, 0, 0, 0x00);
    filldsreq(dsp, (unsigned char *)cue, sz, DSRQ_WRITE | DSRQ_SENSE);
    TIME(dsp) = 120 * 1000; /* Set 2 minute timeout */
    doscsireq(getfd(dsp), dsp);
}

/*
 * A version of read that works the same for std input as files
 */
static int
read_data(int fd, void *buf, int n)
{
    char *ptr = buf;
    int num_read = 0, this_time;

    while (num_read < n && 0 < (this_time = read(fd, ptr, n - num_read))) {
	num_read += this_time;
	ptr += this_time;
    }

    return (num_read);
}

/*
 * Tell the writer to stop
 */
static void
abort_writing(dsreq_t *dsp)
{
    fillg1cmd(dsp, FIXPROTO CMDBUF(dsp), DISCONTINUE,
     0, 0, 0, 0, 0, 0, 0, 0, 0);
    doscsireq(getfd(dsp), dsp);
}

/*
 * Catch signals; abort writing
 */

static void
sig_abort(int sig)
{
    abort_writing(sig_dsp);
    exit(sig);
}

/*
 * Send data to the writer
 */
static void
write_continue(dsreq_t *dsp, unsigned char *dbuf, int count)
{
    fillg1cmd(dsp, FIXPROTO CMDBUF(dsp), WCONT,
     B3(count), 0, 0, 0, 0, 0, 0);
    filldsreq(dsp, dbuf, count, DSRQ_WRITE | DSRQ_SENSE );
    TIME(dsp) = 7200 * 1000; /* Set 120 minute timeout */
    doscsireq(getfd(dsp), dsp);
}

/*
 *  int
 *  mkdisc(char *dev, char *image)
 *
 *  Description:
 *      Make a WORM disc using the Sony CDW-E1 ecoding unit corresponding
 *      to the devscsi device dev.  The data on the disc comes from
 *      the file named by image.
 *
 *  Parameters:
 *      dev         name of devscsi device to use
 *      image       name of file containing image for WORM
 *      max_retries maximum number of times to retry commands
 *      data_size   size of data stream.  This ignored unless image
 *                  is NULL, in which case the data is read from stdin.
 *      verbose     if non-zero, more debbugging output
 *
 *  Returns:
 *      0 if successful, non-zero otherwise
 */
int
mkdisc(char *dev, char *image, int max_retries, int data_size, int verbose)
{
    dsreq_t *dsp;
    int img;
    cuesheet_t cuesheet;
    struct stat st;
    int frames, min, sec, frame, blocks; 
    int bytes_left, count, dbuflen, slop, sz;
    unsigned char *dbuf;
    char maxblklen[6];

    dsp = dsopen(dev, O_RDWR);

    if (!dsp) {
	perror(dev);
	return 1;
    }

    /*
     * Retry because
     *  (1) the device could have been just turned on and
     *  (2) it sometimes needs to be poked a few times
     */
    do {
	testunitready00(dsp);
	if (verbose && STATUS(dsp) != STA_GOOD) {
	    fprintf(stderr, "Device not ready\n");
	    PRINTSENSE(dsp);
	}
    } while (max_retries-- && STATUS(dsp) == STA_CHECK);

    /*
     * Figure out the maximum transfer length for WCONT and malloc
     * a buffer of that size
     */
    fillg0cmd(dsp, FIXPROTO CMDBUF(dsp), RBLIMIT, 0, 0, 0, 0, 0);
    filldsreq(dsp, (unsigned char *)maxblklen, sizeof maxblklen,
     DSRQ_READ | DSRQ_SENSE);
    doscsireq(getfd(dsp), dsp);

    dbuflen = (maxblklen[1] << 16 | maxblklen[2] << 8 | maxblklen[3]);

    dbuf = malloc(dbuflen);
    if (!dbuf) {
	fprintf(stderr, "Out of memory\n");
	return 1;
    }

    if (image) {
	img = open(image, O_RDONLY);

	if (img == -1 || -1 == fstat(img, &st)) {
	    perror(image);
	    return 1;
	}
	sz = st.st_size;
    }
    else {
	img = STDIN_FILENO;
	sz = data_size;
    }

    /*
     * The lead in track
     */
    cuesheet.leadin.control = 0x01;
    cuesheet.leadin.tno = 0;
    cuesheet.leadin.index = 0;
    cuesheet.leadin.dform = 0;
    cuesheet.leadin.zero = 0;
    cuesheet.leadin.min = 0;
    cuesheet.leadin.sec = 0;
    cuesheet.leadin.frame = 0;

    /*
     * Gap between lead in track and first track (actually part
     * of the first track)
     */
    cuesheet.gap.control = 0x61;
    cuesheet.gap.tno = 1;
    cuesheet.gap.index = 0;
    cuesheet.gap.dform = 0x10;
    cuesheet.gap.zero = 0;
    cuesheet.gap.min = 0;
    cuesheet.gap.sec = 0;
    cuesheet.gap.frame = 0;

    /*
     * The data
     */
    cuesheet.data.control = 0x61; /* 0x61 == Data track, copy permitted */
    cuesheet.data.tno = 1;
    cuesheet.data.index = 1;
    cuesheet.data.dform = 0x11; /* 0x11 == generate header/ECC/EDC,
				   use data we send */
    cuesheet.data.zero = 0;
    cuesheet.data.min = 0;
    cuesheet.data.sec = 2;
    cuesheet.data.frame = 0;

    /*
     * Calculate the frame offset of the lead out track.  This will
     * be the number of frames in the data area + 150 frames for
     * the 2 second pause at the beginning of the data track
     * (There are 75 frames/second).
     */
    frames = (sz + CDROM_BLOCKSIZE - 1) / CDROM_BLOCKSIZE + 150;

    /*
     * The lead out track
     */
    cuesheet.leadout.control = 0x01;
    cuesheet.leadout.tno = 0xaa;
    cuesheet.leadout.index = 1;
    cuesheet.leadout.dform = 0;
    cuesheet.leadout.zero = 0;
    min = (frames / 75) / 60;
    cuesheet.leadout.min = INTTOTWODIGITBCD(min);
    sec = (frames / 75) % 60;
    cuesheet.leadout.sec = INTTOTWODIGITBCD(sec);
    frame = frames % 75;
    cuesheet.leadout.frame = INTTOTWODIGITBCD(frame);

    sig_dsp = dsp;
    signal(SIGINT, sig_abort);
    signal(SIGHUP, sig_abort);
    signal(SIGTERM, sig_abort);

    /*
     * Send the cue sheet
     *
     * Retry because of communication problems.
     */
    do {
	write_start(dsp, &cuesheet, sizeof cuesheet);
	if (verbose && STATUS(dsp) == STA_CHECK) {
	    PRINTSENSE(dsp);
	    fprintf(stderr, "Not ready -- retrying\n");
	}
    } while (max_retries-- && STATUS(dsp) == STA_CHECK);

    if (STATUS(dsp) != STA_GOOD) {
	fprintf(stderr, "Error sending cue sheet\n");
	if (verbose)
	    PRINTSENSE(dsp);
	return 1;
    }

    /*
     * We need to add this much at the end to make
     * sure that the amount of data we send is a multiple
     * of CDROM_BLOCKSIZE; otherwise, the writer will think
     * the last frame is only part full.
     */
    slop = (CDROM_BLOCKSIZE - sz % CDROM_BLOCKSIZE) % CDROM_BLOCKSIZE;
    bytes_left = sz;

    printf("%02d:%02d:%02d of data to disc\n", min, sec, frame);
    printf("%d dots until disc is done:\n",
     (bytes_left + CDROM_BLOCKSIZE - 1) / dbuflen);

    while (bytes_left) {
	putchar('.');
	fflush(stdout);
	count = bytes_left > dbuflen ? dbuflen : bytes_left;

	/* Unrecoverable: this trashes the disc */
	if (count != read_data(img, dbuf, count)) {
	    perror(image);
	    abort_writing(dsp);
	    return 1;
	}

	bytes_left -= count;

	if (!bytes_left)
	    bzero(dbuf + count, CDROM_BLOCKSIZE - count % CDROM_BLOCKSIZE);
	
	write_continue(dsp, dbuf, count);

	if (!bytes_left && STATUS(dsp) == STA_CHECK &&
	 SENSEKEY(SENSEBUF(dsp)) == 0 &&
	 ADDSENSECODE(SENSEBUF(dsp)) == 0x80)
	    printf("\nDisc complete");
	else if (STATUS(dsp) != STA_GOOD) {
	    fprintf(stderr, "Error sending data\n");
	    if (verbose)
		PRINTSENSE(dsp);
	    abort_writing(dsp);
	    return 1;
	}
    }

    /*
     * Add some 0 bytes at the end to make it an even multiple
     * of blocksize
     */
    if (slop) {
	unsigned char buf[CDROM_BLOCKSIZE];

	bzero(buf, sizeof buf);
	write_continue(dsp, buf, slop);

	if (STATUS(dsp) == STA_CHECK &&
	 SENSEKEY(SENSEBUF(dsp)) == 0 &&
	 ADDSENSECODE(SENSEBUF(dsp)) == 0x80)
	    printf("\nDisc complete");
	else {
	    fprintf(stderr, "Error sending data\n");
	    if (verbose)
		PRINTSENSE(dsp);
	    return 1;
	}
    }

    putchar('\n');
    return 0;
}
